import io
from pydub import AudioSegment
import logging
from typing import Optional

import requests
from vocode.streaming.agent.bot_sentiment_analyser import BotSentiment
from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.synthesizer import StreamElementsSynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import (
    BaseSynthesizer,
    SynthesisResult,
)


class StreamElementsSynthesizer(BaseSynthesizer):
    TTS_ENDPOINT = "https://api.streamelements.com/kappa/v2/speech"

    def __init__(
        self,
        config: StreamElementsSynthesizerConfig,
        logger: Optional[logging.Logger] = None,
    ):
        super().__init__(config)
        self.voice = config.voice

    def create_speech(
        self,
        message: BaseMessage,
        chunk_size: int,
        bot_sentiment: Optional[BotSentiment] = None,
    ) -> SynthesisResult:
        url_params = {
            "voice": self.voice,
            "text": message.text,
        }
        response = requests.get(self.TTS_ENDPOINT, params=url_params)
        audio_segment: AudioSegment = AudioSegment.from_mp3(
            io.BytesIO(response.content)
        )
        output_bytes_io = io.BytesIO()
        audio_segment.export(output_bytes_io, format="wav")
        return self.create_synthesis_result_from_wav(
            file=output_bytes_io,
            message=message,
            chunk_size=chunk_size,
        )



import asyncio
import json
import logging
import websockets
from websockets.client import WebSocketClientProtocol
import audioop
from urllib.parse import urlencode
from vocode import getenv
import time

from vocode.streaming.transcriber.base_transcriber import (
    BaseTranscriber,
    Transcription,
)
from vocode.streaming.models.transcriber import (
    RevAITranscriberConfig,
    EndpointingConfig,
    EndpointingType,
)
from vocode.streaming.models.audio_encoding import AudioEncoding


NUM_RESTARTS = 5

def getSeconds():
    return time.time()

class RevAITranscriber(BaseTranscriber):
    def __init__(
        self,
        transcriber_config: RevAITranscriberConfig,
        logger: logging.Logger = None,
        api_key: str = None,
    ):
        super().__init__(transcriber_config)
        self.api_key = api_key or getenv("REV_AI_API_KEY")
        if not self.api_key:
            raise Exception(
                "Please set REV_AI_API_KEY environment variable or pass it as a parameter"
            )
        self.transcriber_config = transcriber_config
        self.closed = False
        self.is_ready = True
        self.logger = logger or logging.getLogger(__name__)

    async def ready(self):
        return self.is_ready


    def get_rev_ai_url(self):
        codec = 'audio/x-raw'
        layout = 'interleaved'
        rate = self.transcriber_config.sampling_rate
        audio_format = 'S16LE'
        channels = 1

        content_type = f"{codec};layout={layout};rate={rate};format={audio_format};channels={channels}"

        url_params_dict = {
            "access_token": self.api_key,
            "content_type": content_type,
        }

        url_params_arr = [f'{key}={value}' for (key, value) in url_params_dict.items()]
        url = f"wss://api.rev.ai/speechtotext/v1/stream?" + '&'.join(url_params_arr)
        return url


    async def run(self):
        restarts = 0
        while not self.closed and restarts < NUM_RESTARTS:
            await self.process()
            restarts += 1
            self.logger.debug(
                "Rev AI connection died, restarting, num_restarts: %s", restarts
            )

    async def process(self):
        self.audio_queue = asyncio.Queue()

        async with websockets.connect(self.get_rev_ai_url()) as ws:
            async def sender(ws: WebSocketClientProtocol):
                while not self.closed:
                    try:
                        data = await asyncio.wait_for(self.audio_queue.get(), 5)
                    except asyncio.exceptions.TimeoutError:
                        break
                    await ws.send(data)
                ws.close()
                self.logger.debug("Terminating Rev.AI transcriber sender")

            async def receiver(ws: WebSocketClientProtocol):
                buffer = ""

                while not self.closed:
                    try:
                        msg = await ws.recv()
                    except Exception as e:
                        self.logger.debug(f"Got error {e} in Rev.AI receiver")
                        break
                    data = json.loads(msg)

                    if (data['type'] == 'connected'):
                        continue                    

                    is_done = data['type'] == 'final'
                    if ((len(buffer) > 0)
                        and (self.transcriber_config.endpointing_config)
                        and (self.transcriber_config.endpointing_config.type == EndpointingType.TIME_BASED)
                        and (getSeconds() > self.last_signal_seconds + self.transcriber_config.endpointing_config.time_cutoff_seconds)):
                        is_done = True

                    new_text = ''.join([e['value'] for e in data['elements']])
                    if len(new_text) > len(buffer):
                        self.last_signal_seconds = getSeconds()
                    buffer = new_text

                    confidence = 1.0
                    if is_done:
                        await self.on_response(Transcription(buffer, confidence, True))
                        buffer = ""
                    else:
                        await self.on_response(
                            Transcription(
                                buffer,
                                confidence,
                                False,
                            )
                        )

                self.logger.debug("Terminating Rev.AI transcriber receiver")

            await asyncio.gather(sender(ws), receiver(ws))

    def send_audio(self, chunk):
        self.audio_queue.put_nowait(chunk)

    def terminate(self):
        terminate_msg = json.dumps({"type": "CloseStream"})
        self.audio_queue.put_nowait(terminate_msg)
        self.closed = True



import logging
from typing import Optional
from vocode.streaming.agent.base_agent import BaseAgent
from vocode.streaming.agent.chat_gpt_agent import ChatGPTAgent
from vocode.streaming.agent.echo_agent import EchoAgent
from vocode.streaming.agent.information_retrieval_agent import InformationRetrievalAgent
from vocode.streaming.agent.llm_agent import LLMAgent
from vocode.streaming.agent.restful_user_implemented_agent import (
    RESTfulUserImplementedAgent,
)
from vocode.streaming.models.agent import AgentConfig, AgentType


class AgentFactory:
    def create_agent(
        self, agent_config: AgentConfig, logger: Optional[logging.Logger] = None
    ) -> BaseAgent:
        if agent_config.type == AgentType.LLM:
            return LLMAgent(agent_config=agent_config, logger=logger)
        elif agent_config.type == AgentType.CHAT_GPT:
            return ChatGPTAgent(agent_config=agent_config, logger=logger)
        elif agent_config.type == AgentType.ECHO:
            return EchoAgent(agent_config=agent_config, logger=logger)
        elif agent_config.type == AgentType.INFORMATION_RETRIEVAL:
            return InformationRetrievalAgent(
                agent_config=agent_config,
                logger=logger,
            )
        elif agent_config.type == AgentType.RESTFUL_USER_IMPLEMENTED:
            return RESTfulUserImplementedAgent(agent_config=agent_config, logger=logger)
        raise Exception("Invalid agent config", agent_config.type)



import time
from typing import List
from pydantic import BaseModel, Field
from enum import Enum
from vocode.streaming.models.events import TranscriptEvent

from vocode.streaming.utils.events_manager import EventsManager


class Sender(str, Enum):
    HUMAN = "human"
    BOT = "bot"


class Message(BaseModel):
    text: str
    sender: Sender
    timestamp: float

    def to_string(self, include_timestamp: bool = False) -> str:
        if include_timestamp:
            return f"{self.sender.name}: {self.text} ({self.timestamp})"
        return f"{self.sender.name}: {self.text}"


class Transcript(BaseModel):
    messages: List[Message] = []
    start_time: float = Field(default_factory=time.time)

    def to_string(self, include_timestamps: bool = False) -> str:
        return "\n".join(
            message.to_string(include_timestamp=include_timestamps)
            for message in self.messages
        )

    def add_message(self, text: str, sender: Sender, events_manager: EventsManager, conversation_id: str):
        timestamp = time.time()
        self.messages.append(
            Message(text=text, sender=sender, timestamp=timestamp)
        )
        events_manager.publish_event(
            TranscriptEvent(
                text=text,
                sender=Sender.HUMAN,
                timestamp=time.time(),
                conversation_id=conversation_id,
            )
        )


    def add_human_message(self, text: str, events_manager: EventsManager, conversation_id: str):
        self.add_message(text=text, sender=Sender.HUMAN, events_manager=events_manager, conversation_id=conversation_id)

    def add_bot_message(self, text: str, events_manager: EventsManager, conversation_id: str):
        self.add_message(text=text, sender=Sender.BOT, events_manager=events_manager, conversation_id=conversation_id)



import logging
from typing import Optional
from vocode.streaming.models.transcriber import TranscriberConfig, TranscriberType
from vocode.streaming.transcriber.assembly_ai_transcriber import AssemblyAITranscriber
from vocode.streaming.transcriber.deepgram_transcriber import DeepgramTranscriber
from vocode.streaming.transcriber.google_transcriber import GoogleTranscriber
from vocode.streaming.transcriber.rev_ai_transcriber import RevAITranscriber


class TranscriberFactory:
    def create_transcriber(
        self,
        transcriber_config: TranscriberConfig,
        logger: Optional[logging.Logger] = None,
    ):
        if transcriber_config.type == TranscriberType.DEEPGRAM:
            return DeepgramTranscriber(transcriber_config, logger=logger)
        elif transcriber_config.type == TranscriberType.GOOGLE:
            return GoogleTranscriber(transcriber_config, logger=logger)
        elif transcriber_config.type == TranscriberType.ASSEMBLY_AI:
            return AssemblyAITranscriber(transcriber_config, logger=logger)
        elif transcriber_config.type == TranscriberType.REV_AI:
            return RevAITranscriber(transcriber_config, logger=logger)
        else:
            raise Exception("Invalid transcriber config")



from typing import Optional
from twilio.rest import Client

from vocode.streaming.models.telephony import TwilioConfig


def create_twilio_client(twilio_config: TwilioConfig):
    return Client(twilio_config.account_sid, twilio_config.auth_token)


def end_twilio_call(twilio_client: Client, twilio_sid: str) -> bool:
    response = twilio_client.calls(twilio_sid).update(status="completed")
    return response.status == "completed"



import logging
from pydub import AudioSegment
from typing import Optional
from gtts import gTTS
from io import BytesIO
from vocode.streaming.agent.bot_sentiment_analyser import BotSentiment
from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.synthesizer import GTTSSynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import (
    BaseSynthesizer,
    SynthesisResult,
)


class GTTSSynthesizer(BaseSynthesizer):
    def __init__(
        self,
        config: GTTSSynthesizerConfig,
        logger: Optional[logging.Logger] = None,
    ):
        super().__init__(config)

    def create_speech(
        self,
        message: BaseMessage,
        chunk_size: int,
        bot_sentiment: Optional[BotSentiment] = None,
    ) -> SynthesisResult:
        tts = gTTS(message.text)
        audio_file = BytesIO()
        tts.write_to_fp(audio_file)
        audio_file.seek(0)
        audio_segment: AudioSegment = AudioSegment.from_mp3(audio_file)
        output_bytes_io = BytesIO()
        audio_segment.export(output_bytes_io, format="wav")
        return self.create_synthesis_result_from_wav(
            file=output_bytes_io,
            message=message,
            chunk_size=chunk_size,
        )



import asyncio
from typing import List

from vocode.streaming.models.events import Event, EventType

class EventsManager:

    def __init__(self, subscriptions: List[EventType]):
        self.queue = asyncio.Queue()
        self.subscriptions = set(subscriptions)
        self.active = False

    def publish_event(self, event: Event):
        if event.type in self.subscriptions:
            self.queue.put_nowait(event)

    async def start(self):
        self.active = True
        while self.active:
            try:
                event: Event = await self.queue.get()
            except asyncio.QueueEmpty:
                await asyncio.sleep(1)
            self.handle_event(event)

    def handle_event(self, event: Event):
        pass

    def end(self):
        self.active = False
        while True:
            try:
                event: Event = self.queue.get_nowait()
                self.handle_event(event)
            except asyncio.QueueEmpty:
                break


import asyncio
import ctypes
import io
import pathlib
import wave
from pydub import AudioSegment

import numpy as np
from vocode.streaming.agent.utils import SENTENCE_ENDINGS
from vocode.streaming.models.transcriber import WhisperCPPTranscriberConfig
from vocode.streaming.transcriber.base_transcriber import BaseTranscriber, Transcription
from vocode.utils.whisper_cpp.helpers import transcribe
from vocode.utils.whisper_cpp.whisper_params import WhisperFullParams

WHISPER_CPP_SAMPLING_RATE = 16000


class WhisperCPPTranscriber(BaseTranscriber):
    def __init__(
        self,
        transcriber_config: WhisperCPPTranscriberConfig,
    ):
        super().__init__(transcriber_config)

        self.transcriber_config = transcriber_config
        self._ended = False
        self.buffer_size = round(
            transcriber_config.sampling_rate * transcriber_config.buffer_size_seconds
        )
        self.buffer = np.empty(self.buffer_size, dtype=np.int16)
        self.buffer_index = 0
        self.audio_queue = asyncio.Queue()

        # whisper cpp
        # load library and model
        libname = pathlib.Path().absolute() / self.transcriber_config.libname
        self.whisper = ctypes.CDLL(libname)

        # tell Python what are the return types of the functions
        self.whisper.whisper_init_from_file.restype = ctypes.c_void_p
        self.whisper.whisper_full_default_params.restype = WhisperFullParams
        self.whisper.whisper_full_get_segment_text.restype = ctypes.c_char_p

        # initialize whisper.cpp context
        self.ctx = self.whisper.whisper_init_from_file(
            self.transcriber_config.fname_model.encode("utf-8")
        )

        # get default whisper parameters and adjust as needed
        self.params = self.whisper.whisper_full_default_params()
        self.params.print_realtime = False
        self.params.print_progress = False
        self.params.single_segment = True

    def create_new_buffer(self):
        buffer = io.BytesIO()
        wav = wave.open(buffer, "wb")
        wav.setnchannels(1)
        wav.setsampwidth(2)
        wav.setframerate(self.transcriber_config.sampling_rate)
        return wav, buffer

    async def run(self):
        in_memory_wav, audio_buffer = self.create_new_buffer()
        message_buffer = ""
        while not self._ended:
            chunk = await self.audio_queue.get()
            in_memory_wav.writeframes(chunk)
            if audio_buffer.tell() >= self.buffer_size * 2:
                audio_buffer.seek(0)
                audio_segment = AudioSegment.from_wav(audio_buffer)
                message, confidence = transcribe(
                    self.whisper,
                    self.params,
                    self.ctx,
                    audio_segment,
                )
                message_buffer += message
                is_final = any(
                    message_buffer.endswith(ending) for ending in SENTENCE_ENDINGS
                )
                in_memory_wav, audio_buffer = self.create_new_buffer()
                await self.on_response(
                    Transcription(message_buffer, confidence, is_final)
                )
                if is_final:
                    message_buffer = ""

    def send_audio(self, chunk):
        self.audio_queue.put_nowait(chunk)

    def terminate(self):
        pass
